@@ -521,16 +521,6 @@ tcp_stats(u_long off , char *name, int af )
 	p(tcps_keepdrops, "\t\t%lu connection%s dropped by keepalive\n");
 	p(tcps_predack, "\t%lu correct ACK header prediction%s\n");
 	p(tcps_preddat, "\t%lu correct data packet header prediction%s\n");
-	p(tcps_sack_recovery_episode, "\t%lu SACK recovery episode%s\n"); 
-	p(tcps_sack_rexmits,
-		"\t%lu segment rexmit%s in SACK recovery episodes\n");
-	p(tcps_sack_rexmit_bytes,
-		"\t%lu byte rexmit%s in SACK recovery episodes\n"); 
-	p(tcps_sack_rcv_blocks,
-		"\t%lu SACK option%s (SACK blocks) received\n"); 
-	p(tcps_sack_send_blocks, "\t%lu SACK option%s (SACK blocks) sent\n"); 
-	p1a(tcps_sack_sboverflow, "\t%lu SACK scoreboard overflow\n"); 
-
 #undef p
 #undef p1a
 #undef p2
