@@ -125,7 +125,8 @@ struct host {
 	int refcnt;
 	time_t lastuse;
 	struct sockaddr addr;
-	char name[SM_MAXSTRLEN]; /* name is really variable length and must be last! */
+	char *name;	/* host name provided by client via caller_name */
+	char *revname;	/* host name mapped from addr */
 };
 /* list of hosts we monitor */
 TAILQ_HEAD(hostlst_head, host);
@@ -168,7 +169,8 @@ enum hwlock_status { HW_GRANTED = 0, HW_GRANTED_DUPLICATE,
 
 enum partialfilelock_status { PFL_GRANTED=0, PFL_GRANTED_DUPLICATE, PFL_DENIED,
 			      PFL_NFSDENIED, PFL_NFSBLOCKED, PFL_NFSDENIED_NOLOCK, PFL_NFSRESERR, 
-			      PFL_HWDENIED,  PFL_HWBLOCKED,  PFL_HWDENIED_NOLOCK, PFL_HWRESERR};
+			      PFL_HWDENIED,  PFL_HWBLOCKED,  PFL_HWDENIED_NOLOCK, PFL_HWRESERR,
+			      PFL_HWDENIED_STALEFH, PFL_HWDENIED_READONLY };
 
 enum LFLAGS {LEDGE_LEFT, LEDGE_LBOUNDARY, LEDGE_INSIDE, LEDGE_RBOUNDARY, LEDGE_RIGHT};
 enum RFLAGS {REDGE_LEFT, REDGE_LBOUNDARY, REDGE_INSIDE, REDGE_RBOUNDARY, REDGE_RIGHT};
@@ -232,10 +234,9 @@ enum partialfilelock_status	unlock_partialfilelock(
 void	clear_partialfilelock(const char *hostname);
 enum partialfilelock_status	test_partialfilelock(
     const struct file_lock *fl, struct file_lock **conflicting_fl);
-enum nlm_stats	do_test(struct file_lock *fl,
-    struct file_lock **conflicting_fl);
-enum nlm_stats	do_unlock(struct file_lock *fl);
-enum nlm_stats	do_lock(struct file_lock *fl);
+enum nlm4_stats	do_test(struct file_lock *fl, struct file_lock **conflicting_fl);
+enum nlm4_stats	do_unlock(struct file_lock *fl);
+enum nlm4_stats	do_lock(struct file_lock *fl);
 void	do_clear(const char *hostname);
 
 
@@ -1080,6 +1081,7 @@ lock_hwlock(struct file_lock *fl)
 {
 	struct monfile *imf,*nmf;
 	int lflags, flerror;
+	fhandle_t fh;
 
 	/* Scan to see if filehandle already present */
 	LIST_FOREACH(imf, &monfilelist_head, monfilelist) {
@@ -1114,8 +1116,18 @@ lock_hwlock(struct file_lock *fl)
 		return (HW_RESERR);
 	}
 
+	if (fl->filehandle.n_len > NFS_MAX_FH_SIZE) {
+		debuglog("hwlock: bad fh length %d (from %16s): %32s\n",
+		    fl->filehandle.n_len, fl->client_name, strerror(errno));
+		free(nmf->filehandle.n_bytes);
+		free(nmf);
+		return (HW_STALEFH);
+	}
+	fh.fh_len = fl->filehandle.n_len;
+	bcopy(fl->filehandle.n_bytes, fh.fh_data, fh.fh_len);
+
 	/* XXX: Is O_RDWR always the correct mode? */
-	nmf->fd = fhopen((fhandle_t *)fl->filehandle.n_bytes, O_RDWR);
+	nmf->fd = fhopen(&fh, O_RDWR);
 	if (nmf->fd < 0) {
 		debuglog("fhopen failed (from %16s): %32s\n",
 		    fl->client_name, strerror(errno));
@@ -1384,6 +1396,32 @@ retry_blockingfilelocklist(netobj *fh)
 				do_unlock(ifl);
 				/* ifl is NO LONGER VALID AT THIS POINT */
 			}
+		} else if (pflstatus == PFL_HWDENIED_STALEFH) {
+			/*
+			 * Uh oh...
+			 * It would be nice if we could inform the client of
+			 * this error.  Unfortunately, there's no way to do
+			 * that in the NLM protocol (can't send "granted"
+			 * message with an error and there's no "never going
+			 * to be granted" message).
+			 *
+			 * Since there's no chance of this blocked request ever
+			 * succeeding, we drop the lock request rather than
+			 * needlessly keeping it around just to rot forever in
+			 * the blocked lock list.
+			 *
+			 * Hopefully, if the client is still waiting for the lock,
+			 * they will resend the request (and get an error then).
+			 *
+			 * XXX Note: PFL_HWDENIED_READONLY could potentially
+			 * be handled this way as well, although that would
+			 * only be an issue if a file system changed from
+			 * read-write to read-only out from under a blocked
+			 * lock request, and that's far less likely than a
+			 * file disappearing out from under such a request.
+			 */
+			deallocate_file_lock(ifl);
+			/* ifl is NO LONGER VALID AT THIS POINT */
 		} else {
 			/* Reinsert lock back into same place in blocked list */
 			debuglog("Replacing blocked lock\n");
@@ -1394,9 +1432,10 @@ retry_blockingfilelocklist(netobj *fh)
 				LIST_INSERT_HEAD(&blockedlocklist_head, ifl, nfslocklist);
 		}
 
-		if (pflstatus == PFL_GRANTED || pflstatus == PFL_GRANTED_DUPLICATE) {
-			/* If ifl was permanently removed from the list, (e.g the */
-			/* lock was granted), pfl should remain where it's at. */
+		if (pflstatus == PFL_GRANTED || pflstatus == PFL_GRANTED_DUPLICATE ||
+		    pflstatus == PFL_HWDENIED_STALEFH) {
+			/* If ifl was permanently removed from the list, (e.g it */
+			/* was granted or dropped), pfl should remain where it's at. */
 		} else {
 			/* If ifl was left in the list, (e.g it was reinserted back */
 			/* in place), pfl should simply be moved forward to be ifl */
@@ -1483,7 +1522,7 @@ lock_partialfilelock(struct file_lock *fl)
 				retval = PFL_GRANTED;
 			}
 			if (fl->flags & LOCK_MON)
-				monitor_lock_host_by_name(fl->client_name);
+				monitor_lock_host_by_name(fl->client_name, fl->addr);
 			break;
 		case HW_RESERR:
 			debuglog("HW RESERR\n");
@@ -1493,6 +1532,18 @@ lock_partialfilelock(struct file_lock *fl)
 			debuglog("HW DENIED\n");
 			retval = PFL_HWDENIED;
 			break;
+		case HW_DENIED_NOLOCK:
+			debuglog("HW DENIED NOLOCK\n");
+			retval = PFL_HWDENIED_NOLOCK;
+			break;
+		case HW_STALEFH:
+			debuglog("HW STALE FH\n");
+			retval = PFL_HWDENIED_STALEFH;
+			break;
+		case HW_READONLY:
+			debuglog("HW READONLY\n");
+			retval = PFL_HWDENIED_READONLY;
+			break;
 		default:
 			debuglog("Unmatched hwstatus %d\n",hwstatus);
 			break;
@@ -1610,7 +1661,7 @@ unlock_partialfilelock(const struct file_lock *fl)
 				debuglog("HW duplicate lock failure for left split\n");
 			}
 			if (lfl->flags & LOCK_MON)
-				monitor_lock_host_by_name(lfl->client_name);
+				monitor_lock_host_by_name(lfl->client_name, lfl->addr);
 		}
 
 		if (rfl != NULL) {
@@ -1620,7 +1671,7 @@ unlock_partialfilelock(const struct file_lock *fl)
 				debuglog("HW duplicate lock failure for right split\n");
 			}
 			if (rfl->flags & LOCK_MON)
-				monitor_lock_host_by_name(rfl->client_name);
+				monitor_lock_host_by_name(rfl->client_name, rfl->addr);
 		}
 
 		switch (unlstatus) {
@@ -1725,6 +1776,24 @@ clear_partialfilelock(const char *hostname)
 {
 	struct file_lock *ifl, *nfl;
 	enum partialfilelock_status pfsret;
+	struct host *ihp;
+
+	/*
+	 * Check if the name we got from statd is
+	 * actually one reverse-mapped from the client's
+	 * address.  If so, use the name provided as
+	 * the caller_name in lock requests instead so that
+	 * we can correctly identify the client's locks.
+	 */
+	TAILQ_FOREACH(ihp, &hostlst_head, hostlst) {
+		if (ihp->revname && strncmp(hostname, ihp->revname,
+			SM_MAXSTRLEN) == 0) {
+			hostname = ihp->name;
+			debuglog("Clearing locks for %s (%s)\n",
+				hostname, ihp->revname);
+			break;
+		}
+	}
 
 	/* Clear blocking file lock list */
 	clear_blockingfilelock(hostname);
@@ -1814,11 +1883,11 @@ test_partialfilelock(const struct file_lock *fl,
  * the few return codes which the nlm subsystems wishes to trasmit
  */
 
-enum nlm_stats
+enum nlm4_stats
 do_test(struct file_lock *fl, struct file_lock **conflicting_fl)
 {
 	enum partialfilelock_status pfsret;
-	enum nlm_stats retval;
+	enum nlm4_stats retval;
 
 	debuglog("Entering do_test...\n");
 
@@ -1873,11 +1942,11 @@ do_test(struct file_lock *fl, struct file_lock **conflicting_fl)
  * convinced that this should be abstracted out and bounced up a level
  */
 
-enum nlm_stats
+enum nlm4_stats
 do_lock(struct file_lock *fl)
 {
 	enum partialfilelock_status pfsret;
-	enum nlm_stats retval;
+	enum nlm4_stats retval;
 
 	debuglog("Entering do_lock...\n");
 
@@ -1908,10 +1977,22 @@ do_lock(struct file_lock *fl)
 		break;
 	case PFL_NFSRESERR:
 	case PFL_HWRESERR:
+	case PFL_NFSDENIED_NOLOCK:
+	case PFL_HWDENIED_NOLOCK:
 		debuglog("PFL lock resource alocation fail\n");
 		dump_filelock(fl);
 		retval = (fl->flags & LOCK_V4) ? nlm4_denied_nolocks : nlm_denied_nolocks;
 		break;
+	case PFL_HWDENIED_STALEFH:
+		debuglog("PFL_NFS lock denied STALEFH");
+		dump_filelock(fl);
+		retval = (fl->flags & LOCK_V4) ? nlm4_stale_fh : nlm_denied;
+		break;
+	case PFL_HWDENIED_READONLY:
+		debuglog("PFL_NFS lock denied READONLY");
+		dump_filelock(fl);
+		retval = (fl->flags & LOCK_V4) ? nlm4_rofs : nlm_denied;
+		break;
 	default:
 		debuglog("PFL lock *FAILED*");
 		dump_filelock(fl);
@@ -1924,11 +2005,11 @@ do_lock(struct file_lock *fl)
 	return retval;
 }
 
-enum nlm_stats
+enum nlm4_stats
 do_unlock(struct file_lock *fl)
 {
 	enum partialfilelock_status pfsret;
-	enum nlm_stats retval;
+	enum nlm4_stats retval;
 
 	debuglog("Entering do_unlock...\n");
 	pfsret = unlock_partialfilelock(fl);
@@ -2000,6 +2081,12 @@ testlock(struct nlm4_lock *lock, bool_t exclusive, int flags __unused)
 {
 	struct file_lock test_fl, *conflicting_fl;
 
+	if (lock->fh.n_len > NFS_MAX_FH_SIZE) {
+		debuglog("received fhandle size %d, max size %d",
+		    lock->fh.n_len, NFS_MAX_FH_SIZE);
+		return NULL;
+	}
+
 	bzero(&test_fl, sizeof(test_fl));
 
 	test_fl.filehandle.n_len = lock->fh.n_len;
@@ -2029,11 +2116,11 @@ testlock(struct nlm4_lock *lock, bool_t exclusive, int flags __unused)
  * will do the blocking lock.
  */
 
-enum nlm_stats
+enum nlm4_stats
 getlock(nlm4_lockargs *lckarg, struct svc_req *rqstp, const int flags)
 {
 	struct file_lock *newfl;
-	enum nlm_stats retval;
+	enum nlm4_stats retval;
 
 	debuglog("Entering getlock...\n");
 
@@ -2041,6 +2128,12 @@ getlock(nlm4_lockargs *lckarg, struct svc_req *rqstp, const int flags)
 		return (flags & LOCK_V4) ?
 		    nlm4_denied_grace_period : nlm_denied_grace_period;
 
+	if (lckarg->alock.fh.n_len > NFS_MAX_FH_SIZE) {
+		debuglog("received fhandle size %d, max size %d",
+		    lckarg->alock.fh.n_len, NFS_MAX_FH_SIZE);
+		return (flags & LOCK_V4) ? nlm4_failed : nlm_denied;
+	}
+
 	/* allocate new file_lock for this request */
 	newfl = allocate_file_lock(&lckarg->alock.oh, &lckarg->alock.fh,
 				   (struct sockaddr *)svc_getcaller(rqstp->rq_xprt),
@@ -2052,11 +2145,6 @@ getlock(nlm4_lockargs *lckarg, struct svc_req *rqstp, const int flags)
 		    nlm4_denied_nolocks : nlm_denied_nolocks;
 	}
 
-	if (lckarg->alock.fh.n_len != sizeof(fhandle_t)) {
-		debuglog("recieved fhandle size %d, local size %d",
-		    lckarg->alock.fh.n_len, (int)sizeof(fhandle_t));
-	}
-
 	fill_file_lock(newfl,
 	    lckarg->exclusive, lckarg->alock.svid, lckarg->alock.l_offset,
 	    lckarg->alock.l_len,
@@ -2097,15 +2185,21 @@ getlock(nlm4_lockargs *lckarg, struct svc_req *rqstp, const int flags)
 
 
 /* unlock a filehandle */
-enum nlm_stats
-unlock(nlm4_lock *lock, const int flags __unused)
+enum nlm4_stats
+unlock(nlm4_lock *lock, const int flags)
 {
 	struct file_lock fl;
-	enum nlm_stats err;
-	
-	siglock();
+	enum nlm4_stats err;
 	
 	debuglog("Entering unlock...\n");
+
+	if (lock->fh.n_len > NFS_MAX_FH_SIZE) {
+		debuglog("received fhandle size %d, max size %d",
+		    lock->fh.n_len, NFS_MAX_FH_SIZE);
+		return (flags & LOCK_V4) ? nlm4_failed : nlm_denied;
+	}
+	
+	siglock();
 	
 	bzero(&fl,sizeof(struct file_lock));
 	fl.filehandle.n_len = lock->fh.n_len;
@@ -2123,17 +2217,23 @@ unlock(nlm4_lock *lock, const int flags __unused)
 }
 
 /* cancel a blocked lock request */
-enum nlm_stats
-cancellock(nlm4_cancargs *args, const int flags __unused)
+enum nlm4_stats
+cancellock(nlm4_cancargs *args, const int flags)
 {
 	struct file_lock *ifl, *nfl;
-	enum nlm_stats err;
-
-	siglock();
+	enum nlm4_stats err;
 
 	debuglog("Entering cancellock...\n");
 
-	err = nlm_denied;
+	if (args->alock.fh.n_len > NFS_MAX_FH_SIZE) {
+		debuglog("received fhandle size %d, max size %d",
+		    args->alock.fh.n_len, NFS_MAX_FH_SIZE);
+		return (flags & LOCK_V4) ? nlm4_failed : nlm_denied;
+	}
+
+	siglock();
+
+	err = (flags & LOCK_V4) ? nlm4_denied : nlm_denied;
 
 	/*
 	 * scan blocked lock list for matching request and remove/destroy
@@ -2177,7 +2277,7 @@ cancellock(nlm4_cancargs *args, const int flags __unused)
 		/* got it */
 		remove_blockingfilelock(ifl);
 		deallocate_file_lock(ifl);
-		err = nlm_granted;
+		err = (flags & LOCK_V4) ? nlm4_granted : nlm_granted;
 		break;
 	}
 
@@ -2243,7 +2343,7 @@ get_lock_host(struct hostlst_head *hd, const char *hostname, const struct sockad
  * inform statd
  */
 void
-monitor_lock_host_by_name(const char *hostname)
+monitor_lock_host_by_name(const char *hostname, const struct sockaddr *saddr)
 {
 	struct host *ihp;
 
@@ -2252,11 +2352,15 @@ monitor_lock_host_by_name(const char *hostname)
 	if (ihp == NULL)
 		ihp = get_lock_host(&hostlst_unref, hostname, NULL);
 	if (ihp != NULL) {
-		debuglog("Monitor_lock_host: %s (cached)\n", hostname);
+		if (ihp->revname)
+			debuglog("Monitor_lock_host: %s - %s (cached)\n",
+				ihp->name, ihp->revname);
+		else
+			debuglog("Monitor_lock_host: %s (cached)\n", ihp->name);
 		return;
 	}
 
-	monitor_lock_host(hostname, NULL);
+	monitor_lock_host(hostname, saddr);
 }
 
 void
@@ -2294,13 +2398,15 @@ static void
 monitor_lock_host(const char *hostname, const struct sockaddr *saddr)
 {
 	struct host *nhp;
+	struct hostent *hp = NULL;
 	struct mon smon;
 	struct sm_stat_res sres;
-	int rpcret, statflag;
+	int rpcret;
+	int retrying = 0;
 	size_t n;
+	struct sockaddr_in *sin = (struct sockaddr_in *) saddr;
 
 	rpcret = 0;
-	statflag = 0;
 
 	/* Host is not yet monitored, add it */
 	debuglog("Monitor_lock_host: %s (creating)\n", hostname);
@@ -2309,15 +2415,20 @@ monitor_lock_host(const char *hostname, const struct sockaddr *saddr)
 		debuglog("monitor_lock_host: hostname too long\n");
 		return;
 	}
-	nhp = malloc(sizeof(*nhp) - sizeof(nhp->name) + n + 1);
+	nhp = (struct host *) malloc(sizeof(struct host));
 	if (nhp == NULL) {
 		debuglog("Unable to allocate entry for statd mon\n");
 		return;
 	}
 
 	/* Allocated new host entry, now fill the fields */
-	memcpy(nhp->name, hostname, n);
-	nhp->name[n] = 0;
+	nhp->name = strdup(hostname);
+	if (nhp->name == NULL) {
+		debuglog("Unable to allocate entry name for statd mon\n");
+		free(nhp);
+		return;
+	}
+	nhp->revname = NULL;
 	nhp->refcnt = 1;
 	nhp->lastuse = currsec;
 	if (saddr) {
@@ -2338,27 +2449,51 @@ monitor_lock_host(const char *hostname, const struct sockaddr *saddr)
 	smon.mon_id.my_id.my_vers = NLM_SM;
 	smon.mon_id.my_id.my_proc = NLM_SM_NOTIFY;
 
+retry:
 	rpcret = callrpc("localhost", SM_PROG, SM_VERS, SM_MON, xdr_mon,
 	    &smon, xdr_sm_stat_res, &sres);
 
 	if (rpcret == 0) {
-		if (sres.res_stat == stat_fail) {
+		if (sres.res_stat == stat_fail && !retrying) {
 			debuglog("Statd call failed\n");
-			statflag = 0;
-		} else {
-			statflag = 1;
+			/*
+			 * It's possible that the hostname provided
+			 * by the client isn't valid. Retry with a
+			 * a hostname reverse-mapped from the client's
+			 * address (if provided) and stash this name
+			 * in the host entry so that we can identify it
+			 * with this name later when we ask statd to
+			 * unmonitor it.
+			 */
+			if (saddr)
+				hp = gethostbyaddr((char *) &sin->sin_addr,
+					sizeof(sin->sin_addr), AF_INET);
+			if (hp != NULL && strcmp(nhp->name, hp->h_name) != 0) {
+				debuglog("Statd retry with '%s'\n", hp->h_name);
+				smon.mon_id.mon_name = hp->h_name;
+				nhp->revname = strdup(hp->h_name);
+				if (nhp->revname == NULL) {
+					debuglog("No memory for revname\n");
+					free(nhp->name);
+					free(nhp);
+					return;
+				}
+				retrying = 1;
+				goto retry;
+			}
 		}
 	} else {
 		debuglog("Rpc call to statd failed with return value: %d\n",
 		    rpcret);
-		statflag = 0;
 	}
 
-	if (statflag == 1) {
-		TAILQ_INSERT_HEAD(&hostlst_head, nhp, hostlst);
-	} else {
-		free(nhp);
-	}
+	/*
+	 * Insert host in the monitor list, even if statd
+	 * doesn't like it.  If the name is unacceptable
+	 * to statd, at least we'll avoid subsequent rejection
+	 * on every lock request.
+	 */
+	TAILQ_INSERT_HEAD(&hostlst_head, nhp, hostlst);
 }
 
 /*
@@ -2402,12 +2537,18 @@ destroy_lock_host(struct host *ihp)
 	struct mon_id smon_id;
 	struct sm_stat smstat;
 	int rpcret;
+	char *name;
 
-	debuglog("Attempting to unmonitor host %16s\n", ihp->name);
+	/*
+	 * If the client was monitored with a hostname obtained from
+	 * its address, then use that instead of the possibly invalid
+	 * hostname provided in caller_name.
+	 */
+	name = ihp->revname ? ihp->revname : ihp->name;
+	debuglog("Attempting to unmonitor host %16s\n", name);
 
 	bzero(&smon_id,sizeof(smon_id));
-
-	smon_id.mon_name = (char *)ihp->name;
+	smon_id.mon_name = name;
 	smon_id.my_id.my_name = "localhost";
 	smon_id.my_id.my_prog = NLM_PROG;
 	smon_id.my_id.my_vers = NLM_SM;
@@ -2424,6 +2565,10 @@ destroy_lock_host(struct host *ihp)
 	}
 
 	TAILQ_REMOVE(&hostlst_unref, ihp, hostlst);
+	if (ihp->name)
+		free(ihp->name);
+	if (ihp->revname)
+		free(ihp->revname);
 	free(ihp);
 }
 
@@ -2613,7 +2758,7 @@ granted_failed(nlm4_res *arg)
 /*
  * getshare: try to acquire a share reservation
  */
-enum nlm_stats
+enum nlm4_stats
 getshare(nlm_shareargs *shrarg, struct svc_req *rqstp, const int flags)
 {
 	struct sharefile *shrfile;
@@ -2629,6 +2774,12 @@ getshare(nlm_shareargs *shrarg, struct svc_req *rqstp, const int flags)
 			nlm_denied_grace_period;
 	}
 
+	if (shrarg->share.fh.n_len > NFS_MAX_FH_SIZE) {
+		debuglog("received fhandle size %d, max size %d",
+		    shrarg->share.fh.n_len, NFS_MAX_FH_SIZE);
+		return (flags & LOCK_V4) ? nlm4_failed : nlm_denied;
+	}
+
 	/* find file in list of share files */
 	LIST_FOREACH(shrfile, &nfssharefilelist_head, sharefilelist) {
 		if ((shrarg->share.fh.n_len == shrfile->filehandle.n_len) &&
@@ -2641,8 +2792,11 @@ getshare(nlm_shareargs *shrarg, struct svc_req *rqstp, const int flags)
 
 	/* if share file not found, create a new share file */
 	if (!shrfile) {
+		fhandle_t fh;
 		int fd;
-		fd = fhopen((fhandle_t *)shrarg->share.fh.n_bytes, O_RDONLY);
+		fh.fh_len = shrarg->share.fh.n_len;
+		bcopy(shrarg->share.fh.n_bytes, fh.fh_data, fh.fh_len);
+		fd = fhopen(&fh, O_RDONLY);
 		if (fd < 0) {
 			debuglog("fhopen failed (from %16s): %32s\n",
 			    shrarg->share.caller_name, strerror(errno));
@@ -2659,7 +2813,7 @@ getshare(nlm_shareargs *shrarg, struct svc_req *rqstp, const int flags)
 		if (!shrfile) {
 			debuglog("getshare failed: can't allocate sharefile\n");
 			close(fd);
-			return (flags & LOCK_V4) ? nlm4_failed : nlm_denied;
+			return (flags & LOCK_V4) ? nlm4_denied_nolocks : nlm_denied_nolocks;
 		}
 		shrfile->filehandle.n_len = shrarg->share.fh.n_len;
 		shrfile->filehandle.n_bytes = malloc(shrarg->share.fh.n_len);
@@ -2667,7 +2821,7 @@ getshare(nlm_shareargs *shrarg, struct svc_req *rqstp, const int flags)
 			debuglog("getshare failed: can't allocate sharefile filehandle\n");
 			free(shrfile);
 			close(fd);
-			return (flags & LOCK_V4) ? nlm4_failed : nlm_denied;
+			return (flags & LOCK_V4) ? nlm4_denied_nolocks : nlm_denied_nolocks;
 		}
 		bcopy(shrarg->share.fh.n_bytes, shrfile->filehandle.n_bytes,
 			shrarg->share.fh.n_len);
@@ -2686,13 +2840,13 @@ getshare(nlm_shareargs *shrarg, struct svc_req *rqstp, const int flags)
 			sh->mode = shrarg->share.mode;
 			sh->access = shrarg->share.access;
 			debuglog("getshare: updated existing share\n");
-			return nlm_granted;
+			return (flags & LOCK_V4) ? nlm4_granted : nlm_granted;
 		}
 		if (((shrarg->share.mode & sh->access) != 0) ||
 		    ((shrarg->share.access & sh->mode) != 0)) {
 			/* share request conflicts with existing share */
 			debuglog("getshare: conflicts with existing share\n");
-			return nlm_denied;
+			return (flags & LOCK_V4) ? nlm4_denied : nlm_denied;
 		}
 	}
 
@@ -2712,7 +2866,7 @@ getshare(nlm_shareargs *shrarg, struct svc_req *rqstp, const int flags)
 			free(shrfile->filehandle.n_bytes);
 			free(shrfile);
 		}
-		return (flags & LOCK_V4) ? nlm4_failed : nlm_denied;
+		return (flags & LOCK_V4) ? nlm4_denied_nolocks : nlm_denied_nolocks;
 	}
 	bzero(sh, sizeof(*sh) - sizeof(sh->client_name));
 	sh->oh.n_len = shrarg->share.oh.n_len;
@@ -2726,7 +2880,7 @@ getshare(nlm_shareargs *shrarg, struct svc_req *rqstp, const int flags)
 			free(shrfile->filehandle.n_bytes);
 			free(shrfile);
 		}
-		return (flags & LOCK_V4) ? nlm4_failed : nlm_denied;
+		return (flags & LOCK_V4) ? nlm4_denied_nolocks : nlm_denied_nolocks;
 	}
 	memcpy(sh->client_name, shrarg->share.caller_name, n);
 	sh->client_name[n] = 0;
@@ -2739,19 +2893,25 @@ getshare(nlm_shareargs *shrarg, struct svc_req *rqstp, const int flags)
 
 	debuglog("Exiting getshare...\n");
 
-	return nlm_granted;
+	return (flags & LOCK_V4) ? nlm4_granted : nlm_granted;
 }
 
 
 /* remove a share reservation */
-enum nlm_stats
-unshare(nlm_shareargs *shrarg, struct svc_req *rqstp)
+enum nlm4_stats
+unshare(nlm_shareargs *shrarg, struct svc_req *rqstp, const int flags)
 {
 	struct sharefile *shrfile;
 	struct file_share *sh;
 
 	debuglog("Entering unshare...\n");
 
+	if (shrarg->share.fh.n_len > NFS_MAX_FH_SIZE) {
+		debuglog("received fhandle size %d, max size %d",
+		    shrarg->share.fh.n_len, NFS_MAX_FH_SIZE);
+		return (flags & LOCK_V4) ? nlm4_failed : nlm_denied;
+	}
+
 	/* find file in list of share files */
 	LIST_FOREACH(shrfile, &nfssharefilelist_head, sharefilelist) {
 		if ((shrarg->share.fh.n_len == shrfile->filehandle.n_len) &&
@@ -2765,7 +2925,7 @@ unshare(nlm_shareargs *shrarg, struct svc_req *rqstp)
 	/* if share file not found, return success (per spec) */
 	if (!shrfile) {
 		debuglog("unshare: no such share file\n");
-		return nlm_granted;
+		return (flags & LOCK_V4) ? nlm4_granted : nlm_granted;
 	}
 
 	/* find share */
@@ -2779,7 +2939,7 @@ unshare(nlm_shareargs *shrarg, struct svc_req *rqstp)
 	/* if share not found, return success (per spec) */
 	if (!sh) {
 		debuglog("unshare: no such share\n");
-		return nlm_granted;
+		return (flags & LOCK_V4) ? nlm4_granted : nlm_granted;
 	}
 
 	/* remove share from file and deallocate */
@@ -2799,7 +2959,7 @@ unshare(nlm_shareargs *shrarg, struct svc_req *rqstp)
 
 	debuglog("Exiting unshare...\n");
 
-	return nlm_granted;
+	return (flags & LOCK_V4) ? nlm4_granted : nlm_granted;
 }
 
 /*
